# SPDX-License-Identifier: MIT
# Copyright (c) Hanno Böck
#
# Part of badkeys: https://badkeys.info/

from .rsabias import _bitpct
from .smallfactors import smallfactors


def _checkbits(n, width, bitmask):
    # Skip the first chunk, it may contain larger
    # values due to next prime calculation
    _n = n >> width
    setbits = 0
    zerochunks = 0
    while _n:
        setbits |= _n & bitmask
        _n >>= width

        # check if the current chunk is zero
        if (((1 << width) - 1) & _n) == 0:
            zerochunks += 1

    if zerochunks >= 2:
        # if we see more than 1 zero chunk, we've
        # almost certainly hit a false positive
        return False

    return setbits == 0


def rsapoly(n, e=0):  # noqa: ARG001
    pct = _bitpct(n)
    if pct >= 40:
        return False

    # Detection may not work with implausibly small keys
    if n.bit_length() < 500:
        return False

    lowern = n % (1 << (n.bit_length() // 2))
    lowerpct = _bitpct(lowern)
    if lowerpct >= 40:
        # Probably vanity RSA key, not necessarily insecure,
        # handled by rsabias check
        return False

    if smallfactors(n):
        # If we have small prime factors, report nothing
        # and leave it to the smallfactors module
        return False

    if _checkbits(n, 128, 0xffffffffffff0000):
        # Use funny animal names as placeholders
        # for unidentified vulnerability subtypes
        return {"subtest": "nautilus", "biaspct": pct}
    if _checkbits(n, 32, 0x03ff00000):
        return {"subtest": "completeftp", "biaspct": pct}

    return {"biaspct": pct}



# SPDX-License-Identifier: MIT
# Copyright (c) Hanno Böck
#
# Part of badkeys: https://badkeys.info/

def dsasparse(y, p=0):  # noqa: ARG001
    if p.bit_length() != 1024 or (p >> 1000) != 0x800000 :
        return False
    _p = p >> 168
    _p ^= 1 << 855
    while _p:
        if _p & 0x00fffff0 != 0:
            return False
        _p >>= 32

    return {"detected": True}



__version__ = "0.0.19"
__all__ = ["allchecks", "checkcrt", "checkcsr", "checkdnskey", "checkjwk", "checkprivkey",
           "checkpubkey", "checkrsa", "checksshpubkey", "defaultchecks", "detectandcheck",
           "scanssh", "warningchecks"]
from .checks import (allchecks, checkcrt, checkcsr, checkprivkey, checkpubkey, checkrsa,
                     checksshpubkey, defaultchecks, detectandcheck, warningchecks)
from .dnssec import checkdnskey
from .jwk import checkjwk
from .scanssh import scanssh



# SPDX-License-Identifier: MIT
# Copyright (c) Hanno Böck
#
# Part of badkeys: https://badkeys.info/
#
# Checks for unusual ratio of 0 and 1 bits in RSA modulus


from .smallfactors import smallfactors


def _bitpct(x):
    bitlen = x.bit_length()
    bitset = x.bit_count()
    return bitset * 100 // bitlen


def rsabias(n, e=0):  # noqa: ARG001
    pct = _bitpct(n)

    # For smaller keys, use a slightly larger threshold to
    # avoid false positives
    if n.bit_length() > 2000:
        threshold = 10
    elif n.bit_length() > 1000:
        threshold = 12
    else:
        threshold = 17

    if 50 - threshold < pct <= 50 + threshold:
        return False

    if smallfactors(n):
        # If we have small prime factors, report nothing
        # and leave it to the smallfactors module
        return False

    lowern = n % (1 << (n.bit_length() // 2))
    lowerpct = _bitpct(lowern)

    if pct < 40 and lowerpct < 40:
        # These will be detected by the rsapoly module
        return False

    if 50 - threshold < lowerpct <= 50 + threshold:
        # No bias in the lower half of the modulus is a strong
        # indication of a "vanity" RSA key
        return {"subtest": "vanity", "biaspct": pct}

    return {"biaspct": pct}
